// hash.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: hash.h,v 1.4 1998/10/18 09:54:54 jks Exp $
//

#ifndef HASH_H
#define HASH_H

#include "common.h"

// class hashStringTable
//  A hash implementation.
//
// 

/** Class hashStringTable.

    This class provides a traditional interface for hashing items
    except that the size of the hash table is constant. The items are
    hashed according to their names. If a hash collision occurs, the
    newer item is hashed to the next available place.

    The hashStringTables are needed in assigning global parameters and
    constants data and initial values and in finding records, that are
    referred to by some other record. The simulator class contains two
    hashStringTable instances, one to hold the global parameters and
    the other to hold the global constants. Each generated record
    class has a hashStringTable, from which every record of that type
    can be found by the name of the record.
	
*/

template<class T> class hashStringTable {
protected:
    /// Maximum size of the hash string table.
    const unsigned int HASH_CONST = 100003;

    /** 
    	The names of the hashed entities.
    */
    char *array[HASH_CONST];
    
    /** The pointers to the hashed items.  */
    T *content[HASH_CONST];

    /** A counter that is increased when a new item is added to this
    	 hashStringTable.  */
    int count;
    
    /** The hash key for an entity i is generated according to 
   	its name with the following formula:
   	\TEX{$key_{i} = (\prod_{i=1}^{length(name)} name[i])$} mod HASH-CONST.
	@return The hash key for the given name.
    */
    unsigned int hash(char *string);

    /**
    	The index of the next item in a sequential walk-through of the
	hash table. 
    */
    unsigned int nextItem;
    
public:
    /** The constructor. */
    hashStringTable();

    /** Adds an item to the hash table. 
	@param string The name of the item.
	@param pointer The item to be added to the hash table.
	@return true, if the insert was ok; false, if an item with an
	identical name already existed in the hash */
    bool add(char *string, T *pointer);

    /** Returns the number of items in the hashStringTable. */
    int items() { return count; }

    /** Searches an item whose name matches the name given as
	argument.
    	@param string The name of the item to be searched.
    	@return A pointer to the item. */
    T *get(char *string);

    /** Returns a pointer to the first item in this
        hashStringTable. Using first() and next() all the items stored
        in the hash table can be retrieved in a sequential manner. */
    T *first();

    /** Returns a pointer to the next item in this hashStringTable.
	Using first() and next() all the items stored in the hash
	table can be retrieved in a sequential manner. */
    T *next();
};


//---------------------------------------------------------------------
// hashStringTable::hash
//
// --------------------------------------------------------------------

template<class T>
hashStringTable<T>::hashStringTable()
    : count(0), nextItem(0)
{
    for (unsigned int i = 0; i < HASH_CONST; ++i) {
	array[i] = 0;
	content[i] = 0;
    }
}

//---------------------------------------------------------------------
// hashStringTable::hash
//
// --------------------------------------------------------------------

template<class T>
unsigned int
hashStringTable<T>::hash(char *string)
{
    unsigned int h = 1;
    for (int i = 0; string[i] != 0; i++) h *= string[i];
    return h % HASH_CONST;
}


//---------------------------------------------------------------------
// hashStringTable::add
//
//
//
// --------------------------------------------------------------------

template<class T>
bool
hashStringTable<T>::add(char *string, T *pointer)
{
    ASSERT((unsigned) count < HASH_CONST-1);
    unsigned int h = hash(string);

    if (array[h] == 0) {
	array[h] = new char[strlen(string)+1];
	strcpy(array[h], string);
	content[h] = pointer;
	count++;
	return true;
    } else {
	while (array[h] != 0) {
	    if (strcmp(array[h], string) == 0)
		return false;
	    h++;
	    if (h == HASH_CONST) h = 0;
	}

	array[h] = new char[strlen(string)+1];
	strcpy(array[h], string);
	content[h] = pointer;
	count++;
	return true;
    }
}


//---------------------------------------------------------------------
// hashStringTable::get
//
//
//
// --------------------------------------------------------------------

template<class T>
T *
hashStringTable<T>::get(char *string)
{
    unsigned int h = hash(string);
    if (array[h] == 0) {
	return 0;
    } else {
	while (array[h] != 0) {
	    if (strcmp(array[h], string) == 0)
		return content[h];
	    h++;
	    if (h == HASH_CONST) h = 0;
	}
    }

    return 0;
}


//---------------------------------------------------------------------
// hashStringTable::first
//
//
//
// --------------------------------------------------------------------

template<class T>
T *
hashStringTable<T>::first()
{
    nextItem = 0;
    return next();
}


//---------------------------------------------------------------------
// hashStringTable::next
//
//
//
// --------------------------------------------------------------------

template<class T>
T *
hashStringTable<T>::next()
{
    T *res = 0;

    while (array[nextItem] == 0 && nextItem < HASH_CONST)
	nextItem++;

    if (nextItem < HASH_CONST) {
	res = content[nextItem];
	nextItem++;
    } else
	nextItem = 0;

    return res;
}


#endif // HASH_H

//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// End:                        //
//-----------------------------//
