// main.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: main.h,v 1.12 1999/11/23 08:42:21 jks Exp $
//

#ifndef MAIN_H
#define MAIN_H

#include <vector>
#include <iterator>
#include <algorithm>
#include <functional>
#include <string>
#include <cstdio>
using namespace std;

// Forward definitions
class quantity;
class parameter;
class scanner;

/** 
 * Abstract class delayedOption encapsulates the simulator options that
 * can only be handled after the data have been read in.
 */
class delayedOption {
public:
    /**
     * Calling execute() orders the option to be handled.
     * It must be called for all delayed options in the order they appear
     * on the command line.
     */
    virtual void execute(void) = 0;
    /**
     * Virtual destructor to prevent a hard-to-find programming error later.
     */
    virtual ~delayedOption() {}
};

/**
 * Class delayedIEOption is a common superclass for delayed "-I" and 
 * "-E" options. It contains the common code, since the options are
 * very similar.
 */
class delayedIEOption: public delayedOption {
protected:
    /**
     * Whether this option includes or excludes parameters in/from the output.
     */
    bool _include;
    /**
     * The pattern string as given on the command line: comma separated
     * glob patterns.
     */
    string _patterns;
public:
    /**
     * Mark the options matched by _patterns to be either included or excluded.
     */
    virtual void execute(void);
};

/**
 * Class delayedIncludeOption handles "-I" include options.
 */
class delayedIncludeOption: public delayedIEOption {
public:
    /** Constructor */
    delayedIncludeOption(string patterns) 
    { _patterns = patterns; _include = true; };
};

/**
 * Class delayedExcludeOption handles "-E" exclude options.
 */
class delayedExcludeOption: public delayedIEOption {
public:
    /** Constructor */
    delayedExcludeOption(string patterns)
    { _patterns = patterns; _include = false; };
};

/**
 * Class delayedDataExcludeOption handles "-S" exclude options.
 */
class delayedDataExcludeOption: public delayedOption {
public:
    /** Constructor */
    delayedDataExcludeOption(void) { };
    /**
     * Mark all non-simulatable parameters to be excluded.
     */
    virtual void execute(void);
};

/**
   Class simulator
  */
class simulator {
private:
    /// The single instance of this class
    static simulator *_instance;

protected:
    /// The constructor.
    simulator(int ac, char **av);
    /// The destructor
    ~simulator();
    /// The classic argument count.
    int argc;
    /// The classic argument vector.
    char **argv;

    /// The buffer for standard output
    char *stdoutBuffer;

    /// The name of this program, i.e. argv[0]
    char *progName;
    /// The name of the possible initialization file.
    char *iniFile;
    /// The random generator seed number.
    unsigned int randSeed;
    /// The number of iteration loops to be run.
    int nrOfIterations;
    /** The number of non-printable iteration loops in the beginning
	of the simulation. */
    int skipIterations;
    /** Which iteration to print after skipped iterations. For example 
	value 2 means 'print every 2nd iteration', value 5 means
	'print every 5th iteration' etc. Default is 1. */ 
    int everyNth;
    /// The current iteration number.
    int iteration;
    /** An indicator for a temporary result request. Set by
	sigUsr1Handler when the user sends the process a USR1 signal. */
    bool tmpResultRequested;
    /// The last modified parameter for sigTermHandler.
    parameter *lastModified;
    
    /// The usage message printing function. This function also exits.
    void usage();
    /** Sets either line or block buffering on stdout, with optional 
      buffer size. If size==0, sets stdout unbuffered. */
    void setBuffering(bool lineBuffering, size_t size = BUFSIZ);
    /** Reads the command line options and set flags and arguments
	accordingly. Creates delayed options. */
    void readOptions();
    /** List of options whose handling must be delayed. */
    vector<delayedOption*> *delayedOptions;
    /** Executes the delayed options _and destroys the list_. */
    void executeDelayedOptions(void);

    /** Creates the global variables of the model. This function is
	generated in the model specific part. */
    void createGlobals();
    /** Reads the input data files and create the record structures of
	the model. This function is generated in the model specific part. */
    void readInput(char *file, bool init);
    /** Reads input for global variables. */
    bool globalReadInput(scanner &sc, bool init);

    /** Initializes the global variables of the model. This function is 
	generated in the model specific part. */
    void initGlobals();
    /** Initializes the records of the model. This function is
	generated in the model specific part. */
    void initRecords();

    /** Prints the model name, the compilation date and compiler
	version that generated the simulation program. This function
	is generated in the model specific part. */
    void printModelName(FILE *file);
    /** Prints the model definition which was used to create the
	simulation program. This function is generated in the model
	specific part. */
    void printModelDef(FILE *file);
    /** Prints a data example for the model. This function is generated
	in the model specific part.  */
    void printDataExample(FILE *file);

    /** Prints the command line, which the user used to invoke the
	program as a simulation result comment. Adding the
	command line to the simulation results gives the user at
	least some chance to recall, where the results of this
	simulation came from. */
    void printCommandLine();
    /** Print the header of the simulation result file. */
    void printHeader();
    /** Print a temporary result file after the iteration during which 
	the user had requested it by sending the simulator a USR1
	signal. The result file name is 'tmpresult' and it replaces a
	similarly named file in the simulator working directory. */
    void printTmpResult();

    /**
    	This function does the actual Metropolis-Hastings simulation.
	For details see description of method parameter::MH().
    */
    void simulate();
    
public:    
    /// The simulator starting routine.
    int start();

    int iterationNumber() const { return iteration; }

    /** Supply data for the single instance. The constructor cannot be
        called directly, so this is how main() can give the data to
	the created object. */
    static void supplyData(int argc, char **argv);

    /// Returns the single instance of the simulator object
    static inline simulator *theInstance();
    
    /** Handler for the TERM signal. The current simulation results
	are tried to be saved as fast as possible, and then the
	simulator exits. */ 
    static void sigTermHandler(int s);
    /** Handler for the USR1 signal. The user may request a temporary
	result file by sending the simulator a USR1 signal. This
	functions sets the tmpResultRequested attribute, which causes a 
	temporary result report to be written to a file. */
    static void sigUsr1Handler(int s);
};

#endif // MAIN_H

// Local Variables:
// mode: c++
// c-basic-offset: 4
// case-replace: nil
// make-backup-files: t
// End:
