// mt-rng.h
//
// This file created on August 20, 1998 by Jouni Seppnen, based on 
// Shawn Cokus's file dated March 8, 1998; see
// http://www.math.keio.ac.jp/~matumoto/cokus.c
//
// @(#)$Id: mt-rng.h,v 1.3 1999/01/12 12:58:32 jks Exp $
//
// In the following, "this library" refers to the files mt-rng.h and
// mt-rng.cc.
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as published by
// the Free Software Foundation (either version 2 of the License or, at your
// option, any later version).  This library is distributed in the hope that
// it will be useful, but WITHOUT ANY WARRANTY, without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
// the GNU Library General Public License for more details.  You should have
// received a copy of the GNU Library General Public License along with this
// library; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA 02111-1307, USA.
//
// Copyright notices from the original version:
//
// This version is a recode by Shawn Cokus (Cokus@math.washington.edu)
// on March 8, 1998 of a version by Takuji Nishimura (who had
// suggestions from Topher Cooper and Marc Rieffel in July-August
// 1997).
//
// The code as Shawn received it included the following notice:
//
//   Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.  When
//   you use this, send an e-mail to <matumoto@math.keio.ac.jp> with
//   an appropriate reference to your work.
//
// It would be nice to CC: <Cokus@math.washington.edu> when you write.
//


#ifndef MT_RNG_H
#define MT_RNG_H

#include <stdlib.h>
#include <time.h>

/**
      This class is an adaptation of the 
      Mersenne Twister random number generator
      by Makoto Matsumoto and Takuji Nishimura, as implemented by
      Shawn Cokus. 
      See 
      \URL[the authors' page]{http://www.math.keio.ac.jp/~matumoto/emt.html}
      for more information.

      According to the original authors, the generator
      has the following benefits:
      \begin{itemize}
      \item It is designed with consideration on the flaws of various 
      existing generators. 
      \item Far longer period and far higher order of equidistribution 
      than any other implemented generators.
      \item Fast generation.
      \item Efficient use of memory.
      \end{itemize}

      @author Jouni Seppnen, based on work by 
      Shawn Cokus, Makoto Matsumoto, and Takuji Nishimura.
 */

class MT_RandomNumberGenerator {
private:
    // A comment from the original source:
    //
    // uint32 must be an unsigned integer type capable of holding at least 32
    // bits; exactly 32 should be fastest, but 64 is better on an Alpha with
    // GCC at -O3 optimization so try your options and see what's best for you

    typedef unsigned long uint32;

    static const uint32 N = 624, M = 397, K = 0x9908B0DFU;
    inline uint32 hiBit(uint32 u)  { return u & 0x80000000U; };
    inline uint32 loBit(uint32 u)  { return u & 0x00000001U; };
    inline uint32 loBits(uint32 u) { return u & 0x7FFFFFFFU; };
    inline uint32 mixBits(uint32 u, uint32 v) { return hiBit(u) | loBits(v);};

    uint32 state[N+1];
    uint32 *next;
    int left;

    void MT_RandomNumberGenerator::seedMT(unsigned long seed);
    uint32 reloadMT();
    inline uint32 randomMT();

public:
    /**
         Constructor. Seeds the RNG with the given seed.
	 
	 @param seed The seed for the random number generator.
     */
    MT_RandomNumberGenerator(long seed) {left=-1; setSeed(seed);};
    /**
         Constructor. Seeds the RNG with the time.
     */
    MT_RandomNumberGenerator() {left=-1; setSeed(time(NULL));};
    
    /**
	 Sets the seed of the RNG. The sequence generated by
	 the RNG is repeatable by calling this function again
	 with the same seed.

	 If seed is even, seed+1 is used instead. (The original
	 source says that odd seeds are best.)

         @memo Seed the RNG

	 @param seed The seed for the random number generator.
     */
    void setSeed(unsigned long seed) {
	seedMT(seed);
    }
   
    /**
         Returns a random number uniformly distributed in
	 the specified interval.

         @memo Random number uniformly distributed in [a,b]

	 @param a Lower limit of the interval.
	 @param b Upper limit of the interval.
	 @return A random number.
     */
    double random(double a, double b) {
	return a + (b-a) * 1.0*randomMT()/0xFFFFFFFFU;
    }
};

inline MT_RandomNumberGenerator::uint32 MT_RandomNumberGenerator::randomMT()
{
    register MT_RandomNumberGenerator::uint32 y;
	
    if(--left < 0)
	return(reloadMT());
    
    y  = *next++;
    y ^= (y >> 11);
    y ^= (y <<  7) & 0x9D2C5680U;
    y ^= (y << 15) & 0xEFC60000U;
    return(y ^ (y >> 18));
}

#endif // MT_RNG_H

//------------------------------------------------
// Local Variables:				//
// mode: c++					//
// make-backup-files: t				//
// c-basic-offset: 4                            //
// End:						//
//------------------------------------------------
