// quantity.h
//
// Created 1997 by Karri-Pekka Laakso
//
// @(#)$Id: quantity.h,v 1.9 1999/08/10 08:29:25 jks Exp $
//

#ifndef QUANTITY_H
#define QUANTITY_H

#include <vector>
#include <string>
using namespace std;

#include <stdlib.h>
#include <stdio.h>
#include "common.h"

#include "list.h"

// Forward declarations
class parameter;
class scanner;

/**
   Every named numerical component or value in the simulation inherits
   virtual base class quantity. Every quantity has a value, and
   quantities may be dependent on each other by being parameters
   in each other's distributions. The quantities in the model 
   are divided into three main categories; parameter, constant and
   composite. The latter one is generated by the compiler.
 
   In the simulation (Metropolis-Hastings), 
   when a new value for a quantity is proposed, the algorithm must examine
   the change in the likelihood of the model. Sufficient information 
   to calculate the change of the likelihood is likelihoods of the values 
   that are depending directly on the changed value. This information
   is provided through dependency mechanism that works as follows:
   After all the data has been read every quantity initializes
   its dependency links. Quantities may depend both on the 
   global quantities and on the quantities in their own record.
   When a quantity object q is being initialized, it reports its
   dependence to all the quantities it depends on. This is carried out
   by calling addDependent method of each quantity that quantity q 
   depends on. Quantity q depends on the quantities that are 
   given as parameters to the distribution of q. Parameter is the
   only derived class of quantity (except generated class composite)
   that actually implements this mechanism,
   because parameters are the only quantities whose values may 
   change during the simulation. After the linking every quantity
   that has no initial value must be initialized.
      
   There is a static vector (quantities) of all created quantities.
   This offers an obvious way to find every 
   simulation object in the model, because every simulation object
   in the model inherits class quantity.

   @author Karri-Pekka Laakso
  */
class quantity {
protected:
    /// Name of this quantity
    string name;

    /** Whether the quantity should be printed on the output. */
    bool _printOnOutput;

    /// Global list of all quantities:
    static vector<quantity*> quantities;

public:
    /**
       Constructor.
       Besides creating the quantity, adds the quantity to the list of
       all quantities.  
       
       @param n name of this quantity
    */
    quantity(const string &n);

    /**
       Abstract method that adds a given parameter to the dependency
       list. Used only by the constructor.
       @param *p dependent parameter
      */
    virtual void addDependent(parameter *p) = 0;	

    /**
       Abstract method that returns the current value of this quantity
       @return Current value of this quantity
    */
    virtual double value() const = 0;

    /**
      Should this quantity be printed on the output?
     */
    inline bool printOnOutput() const { return _printOnOutput; }

    /**
      Set whether this quantity should be printed on the output.
      @param  newValue  New behaviour
      @return           Old behaviour
     */
    inline bool printOnOutput(bool newValue) 
    { bool oldValue = _printOnOutput; _printOnOutput = newValue; return oldValue; }

    /**
       Prints the value of this quantity to a given file.
       @param *file target file 
    */
    virtual inline void printValue(FILE *file) const { fprintf(file, "%G ", value()); }

    /**
      Prints the name of this quantity to a given file.
      @param *file target file
    */
    virtual inline void printName (FILE *file) const { fprintf(file," %s", name.c_str()); }

    /**
       Assigns this quantity a data value or an initial value.
       @param &sc scanner that provides mechanisms to read data
       @param *n name of the quantity
       @param init True if the value is an initialization value, false if data.
       @return True if a valid assignment was found, false otherwise.
    */
    virtual bool askValue(scanner &sc, const char *n = 0, bool init = false) = 0;

    /**
       Returns the name of this quantity
       @return Name of this quantity
    */
    inline virtual string identifier() const { return name; };

    /**
       Abstract method that initializes the value of this quantity.
      */
    virtual void initialize() = 0;

    /**
       Static method that returns an iterator to the global vector of all quantities
       @return listIterator to the global vector of all quantities
      */
    static vector<quantity*>::const_iterator begin() { return quantities.begin(); }

    /**
       Static method that returns an iterator to the end of the global vector of all quantities
       @return listIterator to the end of the vector of all quantities
      */
    static vector<quantity*>::const_iterator end() { return quantities.end(); }



    /// Destructor
    virtual ~quantity();
};

#endif // QUANTITY_H

// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:

