// scope.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: scope.cc,v 1.19 1999/08/16 13:13:55 jks Exp $
//

#include "scope.h"
#include "main.h"
#include <string.h>
#include <time.h>

scope *scope::scopes[CG_MAX_SCOPES] = { 0 };
int scope::lastScope = 0;

//---------------------------------------------------------------------
// scope::scope
//  The constructors
//
// The one without arguments in meant for the 0-scope only.
// --------------------------------------------------------------------

scope::scope() :  
    name(""), id(-1), r(0), v(0), echo(true), saveCode(false), 
    codeBuffer("")
{
    if (scopes[0] != 0) {
	printf("Error: an attempt to create a new 0-scope\n");
	c.cleanUp();
	exit(1);
    }
	
    // This is the first scope, which is never deleted (during code
    // generation)

    scopes[0] = this;
    id = 0;
	
    int i;
    for (i=1; i < CG_MAX_SCOPES; ++i)
	scopes[i] = 0;
    
    return;
}


scope::scope(const string &var, const string &group) : 
    name(var), id(-1), r(0), v(0), echo(false), saveCode(true), 
    codeBuffer("")
{
    if (scopes[0] == 0) {
	// What, 0-scope isn't created?! Well, let's create it!
	new scope();
    }

    if (group == "GLOBAL_VAR") {
	v = new XlistIterator<variable>(c.variables->giveIterator());
	setActive();
	return;
    }

    if (group == "GLOBAL_REF") {
	v = new XlistIterator<variable>(c.references->giveIterator());
	setActive();
	return;
    }

    if (group == "RECORD") {
	r = new XlistIterator<record>(c.records->giveIterator());
	setActive();
	return;
    }

    if (group == "COMPOSITE") {
	v = new XlistIterator<variable>(c.composites->giveIterator());
	setActive();
	return;
    }

    if (group == "CONSTANT") {
	v = new XlistIterator<variable>(c.constants->giveIterator());
	setActive();
	return;
    }

    if (group == "PARAMETER") {
	v = new XlistIterator<variable>(c.parameters->giveIterator());
	setActive();
	return;
    }

    if (group == "RECFIELD" || group == "REC_VARFIELD" 
	|| group == "REC_REFFIELD" || group == "REC_OWNVARFIELD") 
	{
	    scope *sc = findScope("REC");
	    if (sc == 0) {
		c.parse_error("Internal error: %s without enclosing REC",
			      group.c_str());

		// Prevent funny errors by creating an empty list
		v = new XlistIterator<variable>(new Xlist<variable>);
		setActive();
		return;
	    }
	
	    XlistIterator<record> r = *(sc->r); 

	    if (r.valid()) {
		// An object of type fieldSelector_t is a pointer to a
		// member function of class record returning Xlist<variable>*.
		typedef Xlist<variable>* (record::* fieldSelector_t)();
		fieldSelector_t fieldSelector;

		if (group == "RECFIELD")
		    fieldSelector = &record::allFields;
		else if (group == "REC_VARFIELD")
		    fieldSelector = &record::variables;
		else if (group == "REC_REFFIELD")
		    fieldSelector = &record::references;
		else // group == "REC_OWNVARFIELD"
		    fieldSelector = &record::ownVariables;

		// l is a pointer to an Xlist of variables. The value
		// of l is obtained by calling fieldSelector on the
		// record pointed to by r. For an explanation of ->*,
		// read Stroustrup chapter 15.5.
		Xlist<variable> *l = ((*r)->*fieldSelector)();
		v = new XlistIterator<variable>(l->giveIterator());
	    } else {
		// r is not valid: return empty list
		v = new XlistIterator<variable>(new Xlist<variable>);
	    }	    

	    setActive();
	    return;
	}
    
    if (group == "PWCFUNCTION") {
	v = new XlistIterator<variable>(new Xlist<variable>); // Empty list at the moment
	setActive();
	return;
    }

    c.parse_error("Internal error: unknown scope %s", group.c_str());

    // Just to prevent further errors:
    v = new XlistIterator<variable>(new Xlist<variable>);
    setActive();
}


//---------------------------------------------------------------------
// scope::setActive
//  Set this scope as the last (= active) scope in the table.
//
// 
// --------------------------------------------------------------------

void
scope::setActive()
{
    ++lastScope;
    if (lastScope >= CG_MAX_SCOPES) {
	printf("Maximum scoping limit (%d) reached\n", CG_MAX_SCOPES-1);
	c.cleanUp();
	exit(1);
    }
    
    scopes[lastScope] = this;
    id = lastScope;
    
    if (validIterator() && (id == 1 || scopes[lastScope-1]->echo == true)) 
	echo = true;
}


//---------------------------------------------------------------------
// scope:findScope
//  Searches for a scope by name from the table of scopes.
//
// 
// --------------------------------------------------------------------

scope *
scope::findScope(const string &name)
{
    scope *s = 0;
    int i;
    for (i=1; scopes[i] != 0 && i < CG_MAX_SCOPES; ++i) {
	if (name == scopes[i]->name) {
	    s = scopes[i];
	    break;
	}
    }

    return s;
}


//---------------------------------------------------------------------
// scope:loopEnd
//  The loop-at-end routine.
//
//  The loop end in the scanned code is reached. No more code is saved,
//  and this scope is at it's end, if there isn't a new variable or
//  record to work on.
// --------------------------------------------------------------------

bool
scope::loopEnd(const string &text) 
{
    // Insert the gathered code in the previous scope, if this was the
    // first loop and if the previous scope is willing to take it 
    // (= it's not yet running from a codeBuffer) 
    if (id > 1 && saveCode && scopes[id-1]->saveCode) { 
	scopes[id-1]->codeBuffer += this->codeBuffer;
	scopes[id-1]->codeBuffer += text;
    }

    saveCode = false;
    nextVar();

    // Is there anything more to do?
    if (echo && validIterator())
	return false;
    else
	return true;
}


//---------------------------------------------------------------------
// scope::insertText
//  Insert the small pieces of code, which are generated by the
//  variables and records.
// --------------------------------------------------------------------

void
scope::insertText(const string &var, const string &text)
{
    scope *sc = findScope(var);

    if (sc == 0) {
	c.parse_error("Internal error: variable %s not found in scope %s",
		      var.c_str(), name.c_str());
	return;
    }

    if (sc->isRecordScope()) {
	record *r = sc->giveRecord();
    
	if (r == 0 || s->echo == false) // Note: s is the global active scope
	    return;
    
	if (text == "NAME") {
	    fprintf(c.file, "%s", r->giveName().c_str());
	    return;
	}

	if (text == "TYPE") {
	    fprintf(c.file, "%s", r->giveType().c_str());
	    return;
	}

	if (text == "BASE") {
	    record *base = r->giveBase();
	    string name;

	    if (base)
		name = base->giveType();
	    else
		name = "record";
	    
	    fprintf(c.file, "%s", name.c_str());
	    return;
	}

    } else {			// ! sc->isRecordScope()
	variable *v = sc->giveVariable();
    
	if (v == 0 || s->echo == false) // Note: s is the global active scope
	    return;
    
	if (text == "NAME") {
	    fprintf(c.file, "%s", v->giveCodeName().c_str());
	    return;
	}

	if (text == "VNAME") {
	    fprintf(c.file, "%s", v->giveName().c_str());
	    return;
	}

	if (text == "TYPE") {
	    fprintf(c.file, "%s", v->giveType().c_str());
	    return;
	}

	if (text == "CREATE") {
	    v->genCreate();
	    return;
	}

	if (text == "POINTER") {
	    v->genPointer();
	    return;
	}

	if (text == "ADD_TO_LIST") {
	    v->genAddToList();
	    return;
	}

	if (text == "DEPENDENCIES") {
	    v->genDependencies();
	    return;
	}

	if (text == "INIT") {
	    v->genInit();
	    return;
	}

	if (text == "INITIALIZER") {
	    ((stochasticVar *)v)->genInitializer();
	    return;
	}

	if (text == "PROPOSER") {
	    ((stochasticVar *)v)->genProposer();
	    return;
	}

	if (text == "PROPOSALDENSITY") {
	    ((stochasticVar *)v)->genProposalDensity();
	    return;
	}

	if (text == "REC_DEFINITION") {
	    v->genRecConsDef();
	    return;
	}

	if (text == "REC_PARAMETER") {
	    v->genRecConsParam();
	    return;
	}

	if (text == "REC_INIT") {
	    v->genRecConsInit();
	    return;
	}

	if (text == "DIST") {
	    ((stochasticVar *)v)->genDistName();
	    return;
	}

	if (text == "DISTPVAL") {
	    ((stochasticVar *)v)->genFuncParams();
	    return;
	}

	if (text == "VALFUNCTION") {
	    ((determVar *)v)->genFunction();
	    return;
	}

	if (text == "DEPFUNCTION") {
	    ((determVar *)v)->genDepFunction();
	    return;
	}

	if (text == "IMPLICIT") {
	    ((determVar *)v)->genImplicit();
	    return;
	}

	if (text == "RECNAME") {
	    ((recordRef *)v)->genRecName();
	    return;
	}

	if (text == "DEPENDENCYHACK_H") {
	    v->genDependencyHackH();
	    return;
	}

	if (text == "DEPENDENCYHACK_C") {
	    v->genDependencyHackC();
	    return;
	}

	if (text == "AG_IMPL_PROTO") {
	    v->genAggregatePrototypes();
	    return;
	}

	if (text == "AG_IMPL_CODE") {
	    v->genAggregateCode();
	    return;
	}
    }

    c.parse_error("Internal error: unknown function %s", text.c_str());
}


//---------------------------------------------------------------------
// scope::~scope
//  The destructor
//
// The scopes should be deleted so that the last ones are deleted
// first. An attempt to do something else is an error.
// --------------------------------------------------------------------

scope::~scope()
{
    if (id != lastScope) {
	printf("Fatal error: an attempt to destroy scope %d/%d\n", id,
	       lastScope);
	c.cleanUp();
	exit(1);
    }

    if (r != 0)
	delete r;
    if (v != 0)
	delete v;

    scopes[lastScope] = 0;
    --lastScope;
}


//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// c-basic-offset: 4           //
// End:                        //
//-----------------------------//
