// variable.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: variable.cc,v 1.25 1999/10/26 10:18:54 jks Exp $
//

#include "variable.h"
#include "main.h"
#include "expr.h"

#include <algorithm>
using namespace std;

//---------------------------------------------------------------------
// variable::variable
//  The base variable class constructor
//
// 
// --------------------------------------------------------------------

variable::variable(const string &n) : 
    name(n), codename("var_"+n), type("constant"), 
    myRecord(c.localRecord), usedAggregates()
{
    if (isRedefinition()) 
	c.parse_error("redefinition of variable '%s'", n.c_str());
}

// variable::createInherited
// almost like a copy constructor -- FIXME: should it be?
// uses new to allocate the return value

variable *
variable::createInherited(record *r) const
{
    ASSERT(myRecord != 0);	// can only be called for record vars
    ASSERT(c.localRecord == 0); // should only get called after parsing
    c.localRecord = r;
    variable *v = new variable(name);
    c.localRecord = 0;

    return v;
}

//---------------------------------------------------------------------
// variable::printCode
//  the code reprinting function.
//
// 
// --------------------------------------------------------------------

void
variable::printCode()
{
    fprintf(c.file, "var %s", name.c_str());
}


//---------------------------------------------------------------------
// variable:isRedefinition
//  Check, if there is a variable called 'name' is this name scope. 
//
// This function is used to avoid redefinition of existing variables.
// --------------------------------------------------------------------

bool
variable::isRedefinition() const
{
    XlistIterator<variable> i;
    
    if (myRecord != 0) 
	// Search for a local variable
	i = myRecord->allFields()->giveIterator();
    else
	// Search for a global variable
	i = c.variables->giveIterator();

    for ( ; i.valid(); ++i) 
	if (name == (*i)->name)
	    return true;
    
    return false;
}


//---------------------------------------------------------------------
// variable::genPointer
//  
//
// 
// --------------------------------------------------------------------

void
variable::genPointer()
{
    // We generate a pointer to quantity, although we know the type of
    // the variable. This facilitates inheritance within the model:
    // for example, stochastics in a base class can be overridden by
    // constants in a derived class.
    fprintf(c.file, "quantity *%s;", codename.c_str());
}


//---------------------------------------------------------------------
// variable::genCreate
//  
//
// 
// --------------------------------------------------------------------

void
variable::genCreate()
{
    fprintf(c.file, "%s = new constant(", codename.c_str());
    if (myRecord == 0)
	// outside records
	fprintf(c.file, "\"%s\");", name.c_str());
    else
	// inside a record Foo: use record's member `name'
	fprintf(c.file, "name + \".%s\");", name.c_str());
}


//---------------------------------------------------------------------
// variable::genAddToList
//  
//
// 
// --------------------------------------------------------------------

void
variable::genAddToList()
{
    fprintf(c.file, "(*globalConstants)[%s->identifier()] = %s;", 
	    codename.c_str(), codename.c_str());
}


//---------------------------------------------------------------------
// variable::genRecConsDef
//  
//
// 
// --------------------------------------------------------------------

void
variable::genRecConsDef()
{
    if (myRecord != 0)
	fprintf(c.file, "record_%s *myRecord;", myRecord->giveName().c_str());
}


//---------------------------------------------------------------------
// variable::genRecConsParam
//  
//
// 
// --------------------------------------------------------------------

void
variable::genRecConsParam()
{
     if (myRecord != 0)
	 fprintf(c.file, ", record_%s *rec", myRecord->giveName().c_str()); 
}


//---------------------------------------------------------------------
// variable::genRecConsInit
//  
//
// 
// --------------------------------------------------------------------

void
variable::genRecConsInit()
{
    if (myRecord != 0)
	fprintf(c.file, ", myRecord(rec)");
}


//---------------------------------------------------------------------
// variable::genDependencyHackH
//  
//
// 
// --------------------------------------------------------------------

void
variable::genDependencyHackH()
{
    if (myRecord == 0 || usedAggregates.empty())
	return;

    fprintf(c.file, 
"    // A terrible hack. Apologies in aggregateFunctionCall::collectVars.\n"
"    class dependencyHack {\n"
"    public:\n"
"	void addDependent(parameter *p);\n"
"    };\n"
"protected:\n"
"    static dependencyHack hack;\n"
"public:\n"
"    static dependencyHack *dependencyAdder() { return &hack; };\n");
}

//---------------------------------------------------------------------
// variable::genDependencyHackC
//  
//
// 
// --------------------------------------------------------------------

void
variable::genDependencyHackC()
{
    if (myRecord == 0 || usedAggregates.empty()) 
	return;

    const char *my_type = type.c_str();
    const char *my_record_type = myRecord->giveType().c_str();

    fprintf(c.file, 
	    "%s::dependencyHack %s::hack;\n"
	    "\n"
	    "void\n"
	    "%s::dependencyHack::addDependent(parameter *p)\n"
	    "{\n"
	    "    vector<%s*> *v = %s::giveInstances();\n"
	    "    for (vector<%s*>::iterator i = v->begin(); i != v->end(); i++)\n"
	    "        (*i)->%s->addDependent(p);\n"
	    "}",
	    my_type, my_type, my_type, 
	    my_record_type, my_record_type, my_record_type,
	    codename.c_str());
}

//----------------------------------------------------------------------
// variable::registerAggregateUse
//----------------------------------------------------------------------

void 
variable::registerAggregateUse(const string &function)
{
    //fprintf(stderr, "registering %s for %s, myRecord=\n", function.c_str(), name.c_str(),
    //    myRecord? myRecord->giveName().c_str() : "(none)");
    usedAggregates.insert(function);
}

//----------------------------------------------------------------------
// variable::genAggregatePrototypes
//----------------------------------------------------------------------

void
variable::genAggregatePrototypes()
{
    pointer_to_binary_function<string, variable*, void> printer =
	ptr_fun(aggregateFunctionCall::printImplementationPrototype);

    for_each(usedAggregates.begin(), usedAggregates.end(), 
		 bind2nd(printer, this));
}

//----------------------------------------------------------------------
// variable::genAggregateCode
//----------------------------------------------------------------------

void
variable::genAggregateCode()
{
    pointer_to_binary_function<string, variable*, void> printer =
	ptr_fun(aggregateFunctionCall::printImplementationCode);

    for_each(usedAggregates.begin(), usedAggregates.end(), 
		 bind2nd(printer, this));
}

//----------------------------------------
// stochasticVar
//----------------------------------------

int stochasticVar::genNumber = 0;

//---------------------------------------------------------------------
// stochasticVar::stochasticVar
//  The constructor of the stochastic, i.e. distributed variables.
//
// 
// --------------------------------------------------------------------

stochasticVar::stochasticVar(const string &n, distributionRef *d, bool pred) :
    variable(n), dist(d), predicted(pred)
{
    // Fix the class name
    if (myRecord == 0) {
	type = name + "_parameter";
    } else {
	type = myRecord->giveName() + "_" + name + "_parameter";
    }

    // Create implicit deterministic variables from those parameters
    // that are not simple variable references but expressions
    // including perhaps multiple variable references and constants.
    XlistIterator<expr> j = d->giveParameters()->giveIterator();
    for (; j.valid(); ++j) {
	if ((*j)->isComposite()) {
	    char *genName = new char[20];
	    sprintf(genName, "implicitGen%d", genNumber++);

	    expr *e = *j;
	    
	    // Put it in the local variable list, if we are in a record
	    if (c.localRecord != 0) 
		c.localRecord->allFields()->append(new determVar(genName, e, true));
	    else
		c.variables->append(new determVar(genName, e, true));

	    // Create a local variable reference instead of the
	    // expression, and put it on the list instead of the
	    // expression 
	    variableRef *vref = new variableRef(genName);
	    j.changeContent(vref);
	}
    }

    initClause = 0;
    propClause = 0;
}

// stochasticVar::createInherited
// almost like a copy constructor -- FIXME: should it be?
// uses new to allocate the return value

variable *
stochasticVar::createInherited(record *r) const
{
    ASSERT(myRecord != 0);	// can only be called for record vars
    ASSERT(c.localRecord == 0); // should only get called after parsing
    c.localRecord = r;
    variable *v = new stochasticVar(name, dist, predicted);
    c.localRecord = 0;

    return v;
}


//---------------------------------------------------------------------
// stochasticVar::addInitClause
//---------------------------------------------------------------------

void
stochasticVar::addInitClause(initializationClause *i = 0)
{
    if (initClause != 0)
	c.parse_error("there already is an initialization clause for this variable");
    if (i == 0)
	i = new randomInitializationClause(dist, this);
    initClause = i;
}

//---------------------------------------------------------------------
// stochasticVar::addProposalClause
//---------------------------------------------------------------------

void
stochasticVar::addProposalClause(proposalClause *p)
{
    if (propClause != 0) 
	c.parse_error("this variable already has a proposal clause");
    if (p == 0)
	p = new proposalClause(dist, this); // use prior
    propClause = p;
}

//---------------------------------------------------------------------
// stochasticVar::findReferences
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::findReferences()
{
    XlistIterator<expr> i = dist->giveParameters()->giveIterator();
    for (; i.valid(); ++i)
	(*i)->findReferences();

    if (initClause != 0)
	initClause->findReferences();
    if (propClause != 0)
	propClause->findReferences();
}


//---------------------------------------------------------------------
// stochasticVar::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
stochasticVar::printCode()
{
    fprintf(c.file, "var %s ~ ", name.c_str());
    dist->printCode();
}


//---------------------------------------------------------------------
// stochasticVar::genDistName
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genDistName()
{
    fprintf(c.file, "%s", dist->giveDistribution()->giveCodename().c_str()); 
}

//---------------------------------------------------------------------
// stochasticVar::genFuncParams
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genFuncParams()
{
    XlistIterator<expr> i = dist->giveParameters()->giveIterator();
    for (; i.valid(); ++i) {
	(*i)->printValue();
	fprintf(c.file, ", ");
    }
}



//---------------------------------------------------------------------
// stochasticVar::genCreate
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genCreate()
{
    fprintf(c.file, "%s = new %s(", codename.c_str(), type.c_str());
    if (myRecord == 0)
	// outside records
	fprintf(c.file, "\"%s\");", name.c_str());
    else
	// inside a record
	fprintf(c.file, "name + \".%s\", this);", name.c_str());
}


//---------------------------------------------------------------------
// stochasticVar::genAddToList
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genAddToList()
{
    fprintf(c.file, 
	    "(*globalParameters)[%s->identifier()] = %s;", 
	    codename.c_str(), codename.c_str());
}


//---------------------------------------------------------------------
// stochasticVar::genDependencies
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genDependencies()
{
    if ( !predicted ) {
	XlistIterator<expr> i = dist->giveParameters()->giveIterator();
	for (; i.valid(); ++i) {
	    // Only add dependencies for non-constants:
	    if (! (*i)->isConst()) {
		// FIXME: The variable can also be made constant in the data file:
		//fprintf(c.file, "\tif (!");
		//(*i)->printCodeName();
		//fprintf(c.file, "->isData())\n\t    ");
		// ...but this doesn't work, since it might not be a variable at all.
		fprintf(c.file, "    ");
		(*i)->printCodeName();
		fprintf(c.file, "->addDependent(static_cast<parameter*>(%s));\n", 
			codename.c_str());
		// FIXME: The use of static_cast here is dangerous, unless we
		// are convinced that this code ONLY gets generated for pointers
		// that point to a subclass of parameter* in the code. This is
		// a valid assumption for stochasticVar, at least for now.

		// An idea for a fix: generate "CAST<parameter*>(%s)", where
		// CAST is a macro that expands to either static_cast (when no
		// run-time type information is available) or to dynamic_cast
		// (for debugging compilations). Add an --enable-debugging
		// option to configure.
	    }
	}
    }
}


//---------------------------------------------------------------------
// stochasticVar::genInit
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genInit()
{
    fprintf(c.file, "%s->initialize();", codename.c_str());
}


//---------------------------------------------------------------------
// stochasticVar::genInitializer
//  
//
// 
// --------------------------------------------------------------------

void
stochasticVar::genInitializer()
{
    if (initClause == 0) {
	if (myRecord == 0)
	    fprintf(stderr, "Warning: no initializer for variable %s, "
		    "initializing from prior\n", name.c_str());
	else
	    fprintf(stderr, "Warning: no initializer for variable %s.%s, "
		    "initializing from prior\n", myRecord->giveName().c_str(), 
		    name.c_str());
	initClause = new randomInitializationClause(dist, this);
	/* FIXME: What if the parameters of dist are not all constant?
	   We might abort, or we might somehow add the restriction that
	   the nonconstants be initialized first.
	   Currently, the user gets a slightly confusing error message. */
    }

    initClause->printInitializer();
}


//---------------------------------------------------------------------
// stochasticVar::genProposer
//---------------------------------------------------------------------

void
stochasticVar::genProposer()
{
    if (propClause == 0) {
	if (myRecord == 0)
	    fprintf(stderr, "Warning: no proposer for variable %s, "
		    "proposing from prior\n", name.c_str());
	else
	    fprintf(stderr, "Warning: no proposer for variable %s.%s, "
		    "proposing from prior\n", myRecord->giveName().c_str(), 
		    name.c_str());
	propClause = new proposalClause(dist, this);
    }

    propClause->printProposer();
}


//---------------------------------------------------------------------
// stochasticVar::genProposalDensity
//---------------------------------------------------------------------

void
stochasticVar::genProposalDensity()
{
    ASSERT(propClause != 0);
    propClause->printProposalDensity();
}


//----------------------------------------
// determVar - deterministic variables
//----------------------------------------



//---------------------------------------------------------------------
// determVar::determVar
//  The constructor of the deterministic (x = a+b) variable class.
//
// 
// --------------------------------------------------------------------

determVar::determVar(const string &n, expr *r, bool i) : 
    variable(n), rule(r), implicit(i)
{
    // If the rule is not a constant, this is a composite
    if (!r->isConst()) {
	if (myRecord == 0) {
	    type = name + "_composite";
	} else {
	    type = myRecord->giveName() + "_" + name + "_composite";
	}
    }
}

// determVar::createInherited
// almost like a copy constructor -- FIXME: should it be?
// uses new to allocate the return value

variable *
determVar::createInherited(record *r) const
{
    ASSERT(myRecord != 0);	// can only be called for record vars
    //ASSERT(!implicit);		// not sensible for implicit vars
    // FIXME: inheriting implicit variables may be dangerous??!!
    ASSERT(c.localRecord == 0); // should only get called after parsing
    c.localRecord = r;
    variable *v = new determVar(name, rule, implicit);
    c.localRecord = 0;

    return v;
}


//---------------------------------------------------------------------
// determVar::printCode
//  The code reprinting function.
//
// 
// --------------------------------------------------------------------

void
determVar::printCode()
{
    fprintf(c.file, "var %s = ", name.c_str());
    rule->printCode();
}


//---------------------------------------------------------------------
// determVar::collectTypes
//  
//
// 
// --------------------------------------------------------------------

void
determVar::collectTypes(Xlist<variable> *c, Xlist<variable> *o,
			Xlist<variable> *)
{
    if (type != "constant")
	c->append(this);  // composite
    else
	o->append(this);  // constant
}


//---------------------------------------------------------------------
// determVar::genFunction
//  
//
// 
// --------------------------------------------------------------------

void
determVar::genFunction()
{
    rule->printValue();
}


//---------------------------------------------------------------------
// determVar::genCreate
//  
//
// 
// --------------------------------------------------------------------

void
determVar::genCreate()
{
    if (type == "constant") {
	fprintf(c.file, "%s = new constant(", codename.c_str());
	if (myRecord == 0)
	    fprintf(c.file, "\"%s\", ", name.c_str());
	else
	    fprintf(c.file, "name + \".%s\", ", name.c_str());
	rule->printValue();
	fprintf(c.file, ");");
    } else { 
	fprintf(c.file, "%s = new %s(", codename.c_str(), type.c_str());
	if (myRecord == 0)
	    fprintf(c.file, "\"%s\");", name.c_str());
	else
	    fprintf(c.file, "name + \".%s\", this);", name.c_str());
    }
}

    
//---------------------------------------------------------------------
// determVar::genAddToList
//  
//
// 
// --------------------------------------------------------------------

void
determVar::genAddToList()
{
    if (type == "constant") {
	fprintf(c.file, 
		"(*globalConstants)[%s->identifier()] = %s;", 
		codename.c_str(), codename.c_str());
    }
}


//---------------------------------------------------------------------
// determVar::genDepFunction
//  
//
// 
// --------------------------------------------------------------------

void
determVar::genDepFunction()
{
    if (type != "constant") {
	if ( ! rule->isConst()) {
	    Xlist<char> *varL = new Xlist<char>;
	    rule->collectVars(varL);

	    XlistIterator<char> i = varL->giveIterator();
	    for (; i.valid(); ++i) {
		fprintf(c.file, "%s->addDependent(p);\n    ", *i);
	    }
	}
    }
}


//---------------------------------------------------------------------
// determVar::genImplicit
//  
//
// 
// --------------------------------------------------------------------

void
determVar::genImplicit()
{
    if (type != "constant" && implicit)
	fprintf(c.file, ", true");
}


//---------------------------------------------------------------------
// determVar::genInit
//  
//
// 
// --------------------------------------------------------------------

void
determVar::genInit()
{
    fprintf(c.file, "%s->initialize();", codename.c_str());
}


// Local Variables:    
// mode: c++           
// make-backup-files: t
// c-basic-offset: 4   
// case-replace: nil   
// End:                
